'' =================================================================================================
''
''   File....... jm_lcd4.spin
''   Purpose.... 4-bit HD44780-compatible LCD driver
''   Author..... Jon "JonnyMac" McPhalen  
''               Copyright (C) 2011-16 Jon McPhalen
''               -- see below for terms of use
''   E-mail.....  
''   Started.... 
''   Updated.... 08 JAN 2016
''
'' =================================================================================================

{ 

  Connections:

    LCD.1  :: Ground
    LCD.2  :: +5v
    LCD.3  :: Contrast; connect to wiper of 10K pot with terminals to +5 and GND
    LCD.4  :: RS
    LCD.5  :: RW  (connect to ground; object is write-only)
    LCD.6  :: E   (use pull-down 4.7K-10K to ground)
    LCD.7  :: DB0 (not used in this driver)
    LCD.8  :: DB1 (not used in this driver)     
    LCD.9  :: DB2 (not used in this driver)
    LCD.10 :: DB3 (not used in this driver)
    LCD.11 :: DB4 
    LCD.12 :: DB5 
    LCD.13 :: DB6 
    LCD.14 :: DB7 
    LCD.15 :: backlight power
    LCD.16 :: backlight ground; connect to transistor circuit controlled by BL pin

}


con { fixed io pins }

  RX1     = 31                                                  ' programming / terminal
  TX1     = 30
  
  SDA     = 29                                                  ' eeprom / i2c
  SCL     = 28


con
  
  CLS     = $01                                                 ' clear the LCD 
  CRSR_HM = $02                                                 ' move cursor home
  CRSR_LF = $10                                                 ' move cursor left 
  CRSR_RT = $14                                                 ' move cursor right 
                                                                  
  CGRAM   = $40                                                 ' character ram
  DDRAM   = $80                                                 ' display ram
                                                                 
  LINE0   = DDRAM | $00                                         ' cursor positions for col 1
  LINE1   = DDRAM | $40                                          
  LINE2   = DDRAM | $14                                          
  LINE3   = DDRAM | $54                                          
                                                                 
   #0, CRSR_NONE, CRSR_ULINE, CRSR_BLINK, CRSR_UBLNK            ' cursor types
                                                                 
  #(-1), DISABLE, OFF, ON                                       ' backlight states
                                                                 

obj

  io : "jm_io_basic"                                            ' essential io routines
                                                                   
                                                                  
var                                                               
                                                                  
  long  ms001                                                   ' cnt ticks per ms
  long  us001                                                   ' cnt ticks per us
                                                                 
  long  inuse                                                   ' io pins have been set
                                                                 
  long  bl                                                       
  long  e                                                        
  long  rs                                                       
  long  db4                                                     ' button 0
  long  db7                                                     ' button 3
                                                                 
  long  dispctrl                                                ' display control bits
                                                                 

pub null

  ' This is not a top-level object

                                                                  
pub start(blpin, epin)                                            

'' Initializes LCD driver in 4-bit mode (multi-line)
'' -- blpin defines the backlight control pin (-1 if not used)
'' -- epin defines start of 8-pin group (see startx)

  startx(blpin, epin+0, epin+1, epin+2)


pub startx(blpin, epin, rspin, db4pin) 

'' Initializes LCD driver in 4-bit mode (multi-line)
'' -- allows flexible pin assigments
''    ** DB7..DB4 pins __MUST__ be contiguous group with DB4 as LSB

  ms001 := clkfreq / 1_000
  us001 := clkfreq / 1_000_000

  stop                                                          ' clear existing pins
                                                                 
  bl   := blpin                                                  
  e    := epin                                                  ' lcd.6
  rs   := rspin                                                 ' lcd.4
  db4  := db4pin                                                ' lcd.11
  db7  := db4pin + 3                                            ' lcd.14
                                                                 
  if (bl => 0)                                                  ' backlight used?
    backlight(OFF)                                              '  yes, start off
                                                                 
  lcd_init                                                      ' initialize in 4-bit mode
  inuse := true                                                 ' mark initialized
                                                                 

pub stop

'' Makes LCD buss pins inputs
'' -- works only if LCD has been previously defined

  if (inuse)
    backlight(DISABLE)
    dira[e] := 0    
    dira[rs] := 0      
    dira[db7..db4] := %0000 
    inuse := false


pub clear

'' Clear the LCD

  cmd(CLS)


pub home

'' Move cursor to 0, 0 in display

  cmd(CRSR_HM)


pub left

'' Move cursor one character left

  cmd(CRSR_LF)


pub right

'' Move cursor on character right

  cmd(CRSR_RT)


pub line(n)

'' Move to beginning of specified line

  case n
    0: cmd(LINE0)
    1: cmd(LINE1)
    2: cmd(LINE2)
    3: cmd(LINE3)


pub cmd(c)

'' Write command byte to LCD

  outa[rs] := 0                                                 ' command mode
  wr_lcd(c)
  

pub out(c)

'' Print character byte to LCD

  outa[rs] := 1                                                 ' data mode 
  wr_lcd(c)


pub outx(c, n)

'' Print character n times

  repeat n
    out(c)
      

pub str(p_zstr)

'' Print z-string at p_zstr
'' -- p_zstr is pointer (hub address) to z-string

  repeat strsize(p_zstr)
    out(byte[p_zstr++])


pub substr(p_zstr, n) | c

'' Prints n characters of string at p_zstr
'' -- aborts if end of string detected

  repeat n
    c := byte[p_zstr++]
    if (c > 0) 
      out(c)
    else
      quit


pub nstr(p_zstr, n)

'' Prints nth string from list of z-strings

  repeat while (n > 0)
    p_zstr += (strsize(p_zstr) + 1)                             ' advance to next
    --n

  str(p_zstr)

 
pub dec(value) | i, x

'' Print a decimal number
'  -- borrowed from FullDuplexSerial
                                                                             
  x := (value == negx)                                                         
  if (value < 0)
    value := ||(value+x)                                                     
    out("-")

  i := 1_000_000_000                                                         

  repeat 10                                                                  
    if value => i                                                            
      out(value / i + "0" + x*(i == 1))                                      
      value //= i                                                            
      result~~                                                               
    elseif result or (i == 1)
      out("0")                                                               
    i /= 10  

    
pub rjdec(val, width, pchar) | tmpval, padwidth

'' Print right-justified decimal value
'' -- val is value to print
'' -- width is width of (space padded) field for value

'  Original code by Dave Hein
'  Modifications by Jon McPhalen

  if (val => 0)                                                 ' if positive
    tmpval := val                                               '  copy value
    padwidth := width - 1                                       '  make room for 1 digit
  else                                                           
    if (val == negx)                                            '  if max negative
      tmpval := posx                                            '    use max positive for width
    else                                                        '  else
      tmpval := -val                                            '    make positive
    padwidth := width - 2                                       '  make room for sign and 1 digit
                                                                 
  repeat while (tmpval => 10)                                   ' adjust pad for value width > 1
    padwidth--                                                   
    tmpval /= 10                                                 
                                                                 
  repeat padwidth                                               ' print pad
    out(pchar)                                                   
                                                                 
  dec(val)                                                      ' print value
                                                                 
 
pub hex(value, digits)

'' Print a hexadecimal number
'  -- borrowed from FullDuplexSerial

  digits := 1 #> digits <# 8

  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))


pub bin(value, digits)

'' Print a binary number
'  -- borrowed from FullDuplexSerial

  digits := 1 #> digits <# 32

  value <<= (32 - digits)
  repeat digits
    out((value <-= 1) & 1 + "0")


pub set_char(c, p_def)

'' Write character map data to CGRAM
'' -- c is the custom character # (0..7)
'' -- p_def is pointer to character definition bytes (8)

  if ((c => 0) and (c =< 7))                                    ' legal char # (0..7)?
    cmd(CGRAM + (c << 3))                                       ' move cursor
    repeat 8                                                    ' output character data
      out(byte[p_def++])                                         
    return true                                                  
                                                                 
  else                                                           
    return false                                                 
                                                                 
                                                                 
pub display(ison)                                                
                                                                 
'' Enables/disables display                                      
                                                                 
  if (ison)                                                      
    dispctrl := dispctrl |  %0000_0100                          ' display bit on
  else                                                           
    dispctrl := dispctrl & !%0000_0100                          ' display bit off
                                                                 
  cmd(dispctrl)
  

pub cursor(mode)

'' Sets LCD cursor style: off (0), underline (1), blinking bkg (2), uline+bkg (3)

  case mode
    CRSR_NONE  : dispctrl := dispctrl & %0000_1100 | %0000_1000  
    CRSR_ULINE : dispctrl := dispctrl & %0000_1100 | %0000_1010 
    CRSR_BLINK : dispctrl := dispctrl & %0000_1100 | %0000_1001
    CRSR_UBLNK : dispctrl := dispctrl & %0000_1100 | %0000_1011
    other      : return false

  cmd(dispctrl) 
    
  return true

    
pub crsr_xy(x, y) 

'' Moves DDRAM cursor to column, row position
'' -- home position is indexed as 0, 0

  case y
    0 : cmd(LINE0 + x)
    1 : cmd(LINE1 + x)
    2 : cmd(LINE2 + x)
    3 : cmd(LINE3 + x)


pub wr_mem(addr, p_src, n)

'' Writes n bytes from p_src to addr in display

  cmd(addr)                                                     ' setup where to write
  repeat n                                                       
    out(byte[p_src++])                                           
                                                                 
                                                                 
pub backlight(state)                                             
                                                                 
'' Set backlight control pin                                     
'' -- forces bl pin to output state when ON or OFF               
                                                                 
  if (bl => 0)                                                  ' if used
    if (state > 0)                                              ' on if positive
      io.high(bl)                                                   
    elseif (state == 0)                                         ' off if zero
      io.low(bl)                                                    
    else                                                        ' disable
      io.input(bl)                                                  
                                                                 

pub read_buss

'' Reads state of LCD buss pins
'' -- useful when shared with buttons

  dira[db7..db4] := %0000

  return ina[db7..db4]
                                                            
                                                                  
pri lcd_init                                                      
                                                                  
' Initializes LCD using 4-bit interface                           
                                                                  
  io.low(e)                                                     ' setup pins
  io.low(rs)                                                     
  outa[db7..db4] := %0000                                        
  dira[db7..db4] := %1111                                        
                                                                 
  waitcnt((50 * ms001) + cnt)                                   ' let LCD power up
  outa[db7..db4] := %0011                                       ' 8-bit mode
  blip_e                                                         
                                                                 
  waitcnt((5 * ms001) + cnt)                                     
  blip_e                                                         
                                                                 
  waitcnt((150 * us001) + cnt)                                   
  blip_e                                                         
                                                                 
  outa[db7..db4] := %0010                                       ' 4-bit mode
  blip_e                                                         
                                                                 
  cmd(%0010_1000)                                               ' multi-line
  cmd(%0000_0110)                                               ' auto-increment cursor
  dispctrl := %0000_1100                                        ' display on, no cursor
  cmd(dispctrl)                                                  
                                                                  

pri wr_lcd(b)

' Writes byte b to LCD buss using 4-bit interface
' -- sets LCD buss pins to outputs

  dira[db7..db4] := %1111                                       ' reset LCD buss to outputs
                                                                 
  outa[db7..db4] := b >> 4                                      ' output high nibble                             
  blip_e                                                         
  outa[db7..db4] := b & $0F                                     ' output low nibble
  blip_e                                                         
                                                                 
  ' with no busy flag we must insert time pad                    
  ' -- pads are longer than required by spec for safety          
                                                                 
  if ((b == CLS) or (b == CRSR_HM))                                 
    waitcnt((ms001 << 1) + cnt)                                 ' use 2ms for CLS and HOME
  else                                                           
    waitcnt((ms001 >> 4) + cnt)                                 ' 62.5us for other bytes
                                                                 

pri blip_e

' "Blips" LCD.E pin to transfed data from buss to LCD

  io.high(e)
  io.low(e)


dat { license }

{{

  Copyright (C) 2011-16 Jon McPhalen

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}                   